/*
 * yasapi_engine_client.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int EngineClientPreOpen(EngineClient *pClient, int bIncBroken,
    int bIncTracks);
static EEngineServer EngineClientServerState(EngineClient *pClient);

///////////////////////////////////////////////////////////////////////////////
static void EngineClientUpdateWritten(EngineClient *pClient)
{
  Ring *pRing=&pClient->ring;
  double xWritten=pRing->frames._uWritten;
  double xSize=pRing->frames._uWritten+pRing->frames._uAvailable;

  pClient->monitor.xTailCur=xWritten/xSize;

  // xTailCurMin:
  if (YASAPI_DOUBLE_INVALID(pClient->monitor.xTailMin))
    pClient->monitor.xTailMin=pClient->monitor.xTailCur;
  else if (pClient->monitor.xTailCur<pClient->monitor.xTailMin)
    pClient->monitor.xTailMin=pClient->monitor.xTailCur;

  // xTailCurMax:
  if (YASAPI_DOUBLE_INVALID(pClient->monitor.xTailMax))
    pClient->monitor.xTailMax=pClient->monitor.xTailCur;
  else if (pClient->monitor.xTailMax<pClient->monitor.xTailCur)
    pClient->monitor.xTailMax=pClient->monitor.xTailCur;
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientCreate(EngineClient *pClient, Player *pPlayer)
{
  /////////////////////////////////////////////////////////////////////////////
  memset(pClient,0,sizeof *pClient);
  FsmCreate(pClient,gcaIEngineClientState);

  /////////////////////////////////////////////////////////////////////////////
  pClient->pPlayer=pPlayer;
  EngineMonitorReset(&pClient->monitor);

  return 0;
}

void EngineClientDestroy(EngineClient *pClient)
{
  PlayerLock(pClient->pPlayer ); // {
  FsmDestroy(pClient);
  PlayerUnlock(pClient->pPlayer); // }
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientRequestServerStateEx(EngineClient *pClient, EEngineServer eRequest,
    EEngineServer eReply )
{
  int code=-1;
  EngineServer *pServer=pClient->pServer;
  Player *pPlayer=pClient->pPlayer;
  int nLockRecursion;
  EEngineServer eEngineServer;
  DWORD dwWait;

  DASSERT(pServer,nolock);
  PlayerLock(pPlayer ); // {
  pServer->thread.eStateRequest=eRequest;
DVWRITELNEX(file,line,func,"### %s {",FsmStateIdEx(pServer,eRequest));
  SetEvent(pServer->thread.hEvent);

  for (;;) {
    eEngineServer=FsmState(pServer);

    if (eEngineServer==eReply)
      break;
    else if (eEngineServerStateError==eReply)
      goto exit;

    ResetEvent(pClient->thread.hEvent);
    PlayerUnlockDeep(pPlayer,&nLockRecursion); // {
DWRITELNEX(file,line,func,"{");
    dwWait=WaitForSingleObject(pClient->thread.hEvent,INFINITE);
DWRITELNEX(file,line,func,"}");
    //dwWait=WaitForSingleObject(pClient->thread.hEvent,1000ul);
    PlayerLockDeep(pPlayer,nLockRecursion ); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DMESSAGE("WAIT_ABANDONED");
      goto exit;
    case WAIT_TIMEOUT:
      DMESSAGE("WAIT_TIMEOUT");
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED: %lu",GetLastError());
      goto exit;
    default:
      break;
    }
  }

DVWRITELNEX(file,line,func,"### %s }",FsmStateIdEx(pServer,eRequest));
  code=0;
exit:
  PlayerUnlock(pPlayer); // }
DLABEL(nolock);
  return code;
}

int EngineClientRequestServerState(EngineClient *pClient, EEngineServer eEngineServer )
{
  return EngineClientRequestServerStateEx(pClient,eEngineServer,eEngineServer );
}

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_START_DEBUG
int EngineClientOpen(EngineClient *pClient, int samplerate, int numchannels,
    int bitspersamp, int *pnMaxLatency, int *pnFramesDevice)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  HANDLE hEvent=pClient->thread.hEvent;
  GtkClient *pGtk=&pPlayer->gtk;
  int bIncGaplessBroken=pClient->options.common.bGapless;
  EngineOptions options;
  int bOptionsChanged;
  EngineWfx wfx;
  int bTargetChanged;
  int nLockRecursion;

  ///////////////////////////////////////////////////////////////////////////
  PlayerLock(pPlayer ); // {

  /////////////////////////////////////////////////////////////////////////////
  if (FsmUp(pClient,eEngineClientStateBase,__FILE__,__LINE__,__func__)<0) {
    DMESSAGE("creating producer base");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (EngineOptionsCreate(&options,pClient)<0) {
    DMESSAGE("creating options");
    goto unlock;
  }

  bOptionsChanged=memcmp(&pClient->options,&options,
      sizeof pClient->options);

  /////////////////////////////////////////////////////////////////////////////
  if (EngineWfxCreate(&wfx,&options,samplerate,numchannels,bitspersamp)<0) {
    DMESSAGE("creating WAVEFORMATEXTENSIBLEs");
    goto unlock;
  }

  bTargetChanged=memcmp(&pClient->wfx.target.Format,&wfx.target.Format,
      sizeof pClient->wfx.target.Format);

  /////////////////////////////////////////////////////////////////////////////
  if (!options.common.bGapless) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN("!! !bGapless !!");
#endif // ]
    goto start;
  }

  if (bOptionsChanged) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN("!! bOptionsChanged !!");
#endif // ]
    goto start;
  }

  if (bTargetChanged) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN("!! bTargetChanged !!");
#endif // ]
    goto start;
  }

  if (eEngineClientStateBase==FsmState(pClient)) {
#if defined (YASAPI_START_DEBUG) // [
    DWRITELN("!! eEngineClientStateBase !!");
#endif // ]
    goto start;
  }

  // Continue Gaplessly ///////////////////////////////////////////////////////
  FsmUp(pClient,eEngineClientStateOpen,__FILE__,__LINE__,__func__);
  ++pClient->pPlayer->stats.uNumTracks;
  ++pClient->stats.uNumTracks;
  EngineClientUpdateWritten(pClient);

  if (EngineClientRequestServerState(pClient,eEngineServerStateOpen )<0) {
    DMESSAGE("requesting eEngineServerStateOpen");
    goto exit;
  }

  FsmDown(pClient,eEngineClientStateOpen );

  if (eEngineServerStateOpen<=FsmState(pClient->pServer))
     pClient->param.time.lpVtbl->Inc(pClient->pServer);

  if (RingResetWrite(&pClient->ring,0)<0) {
    DMESSAGE("resetting ring write");
    goto exit;
  }

  memcpy(&pClient->wfx.source,&wfx.source,sizeof pClient->wfx.source);
  goto success;
start:
DVWRITELN("%s ============================",FsmStateId(pClient));
  // Start ////////////////////////////////////////////////////////////////////
  while (eEngineServerStateBase!=FsmState(pClient->pServer)) {
    ResetEvent(hEvent);
    PlayerUnlockDeep(pPlayer,&nLockRecursion); // {
    WaitForSingleObject(hEvent,INFINITE);
    PlayerLockDeep(pPlayer,nLockRecursion ); // }
  }

  memcpy(&pClient->options,&options,sizeof pClient->options);
  memcpy(&pClient->wfx,&wfx,sizeof pClient->wfx);

  ///////////////////////////////////////////////////////////////////////////
DVWRITELN("bIncGaplessBroken: %d",bIncGaplessBroken);
  bIncGaplessBroken=bIncGaplessBroken&&!bTargetChanged;
DVWRITELN("bIncGaplessBroken: %d",bIncGaplessBroken);
  bIncGaplessBroken=bIncGaplessBroken&&!bOptionsChanged;
DVWRITELN("bIncGaplessBroken: %d",bIncGaplessBroken);
  //bIncGaplessBroken=bIncGaplessBroken&&eFsmBase<pFsm->eState;
  bIncGaplessBroken=bIncGaplessBroken&&FsmState(pClient)<eEngineClientStateClose;
DVWRITELN("bIncGaplessBroken: %d",bIncGaplessBroken);

  if (EngineClientPreOpen(pClient,bIncGaplessBroken,1)<0) {
    DMESSAGE("creating gapless sequence");
    goto exit;
  }
success:
  GtkClientUpdateDialog(pGtk,&pGtk->config);

  pClient->param.nMaxLatency=2.0*pClient->options.device.xMaxLatencyAssumed+0.5;

  if (pnMaxLatency)
    *pnMaxLatency=pClient->param.nMaxLatency;

  if (pnFramesDevice)
    *pnFramesDevice=pClient->param.uNumFramesDevice;

  code=0;
unlock:
  PlayerUnlock(pPlayer); // }
exit:
  return code;
}

static void _EngineClientCloseInternal(EngineClient *pClient)
{
DVWRITELN("2 %s {",FsmStateIdEx(pClient,eEngineClientStateBase));
  FsmDown(pClient,eEngineClientStateBase );
DVWRITELN("2 %s }",FsmStateIdEx(pClient,eEngineClientStateBase));
  memset(&pClient->wfx.source,0,sizeof pClient->wfx.source);
  memset(&pClient->wfx.target,0,sizeof pClient->wfx.target);
  memset(&pClient->param,0,sizeof pClient->param);
}

static void _EngineClientClose(EngineClient *pClient)
{
  EEngineClient eEngineClient;
  EEngineServer eEngineServer;
  int bKill=0;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DMESSAGE("not initialized");
    goto enoinit;
  }

  DVWRITELN("=== target: %u/%u/%u ===",
      pClient->wfx.target.Format.nSamplesPerSec,
      pClient->wfx.target.Format.nChannels,
      pClient->wfx.target.Format.wBitsPerSample);

  eEngineClient=FsmState(pClient);

  if (eEngineClientStateOpen<=eEngineClient) {
    eEngineServer=EngineClientServerState(pClient);

    DVWRITELN("bKill: %d",bKill);
    bKill=bKill||!pClient->param.bActive;
DVWRITELN("bKill: %d (!pClient->param.bActive)",bKill);
    bKill=bKill||eEngineServer<eEngineServerStateDraining;
DVWRITELN("bKill: %d (%s<eEngineServerStateDraining)",bKill,FsmStateIdEx(pClient->pServer,eEngineServer));
    bKill=bKill||!pClient->options.common.bGapless;
DVWRITELN("bKill: %d (!pClient->options.common.bGapless)",bKill);

    if (bKill) {
DWRITELN("=== KILLING ==================================================");
      pClient->param.bActive=0;
      _EngineClientCloseInternal(pClient);
      FsmUp(pClient,eEngineClientStateClose,__FILE__,__LINE__,__func__);
    }
  }

  DVWRITELN("=== target: %u/%u/%u ===",
      pClient->wfx.target.Format.nSamplesPerSec,
      pClient->wfx.target.Format.nChannels,
      pClient->wfx.target.Format.wBitsPerSample);
enoinit: ;
DVWRITELN("%s ============================================",FsmStateId(pClient));
}

void EngineClientClose(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;

DVWRITELN("### %s ####################################",__func__);
  if (FsmState(pClient)<eEngineClientStateBase) {
    DMESSAGE("not initialized");
    goto enoinit;
  }

  PlayerLock(pPlayer ); // {
DWRITELN(FsmStateId(pClient));

  if (eEngineClientStateClose<FsmState(pClient))
    _EngineClientClose(pClient);

  PlayerUnlock(pPlayer); // }
enoinit: ;
DVWRITELN("%s ============================================",FsmStateId(pClient));
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientAvailable(EngineClient *pClient)
{
  Ring *pRing=&pClient->ring;
  int nAvailable=0;

  if (FsmState(pClient)<eEngineClientStateRing)
    nAvailable=INT_MAX;
  else {
    switch (FsmState(pClient->pServer)) {
    case eEngineServerStatePauseConnected:
    case eEngineServerStatePauseDisconnected:
      break;
    default:
      nAvailable=RingSourceSizeAvailable(pRing);
      break;
    }
  }

  return nAvailable;
}

int EngineClientAvailable(EngineClient *pClient)
{
  int nAvailable=0;
  Player *pPlayer=pClient->pPlayer;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  PlayerLock(pPlayer ); // {
  nAvailable=_EngineClientAvailable(pClient);
  PlayerUnlock(pPlayer); // }
ebase:
  return nAvailable;
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientWrite(EngineClient *pClient, const BYTE *buf, DWORD dwChunkSizeSource)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;
  GtkClient *pGtk=&pPlayer->gtk;
  Ring *pRing=&pClient->ring;
  WORD uFramesPlaybackBegin;
  WORD uFramesWritten;
  WORD uFramesMax;
  UINT32 uMsLatency;
  EEngineClient eEngineClient;

  eEngineClient=FsmState(pClient);

  if (eEngineClient<eEngineClientStateBase) {
    DMESSAGE("producer not initialized");
    goto ebase;
  }
  else if (eEngineClientStateBase==eEngineClient) {
    if (EngineClientPreOpen(pClient,1,0)<0) {
      DMESSAGE("creating gapless sequence");
      goto exit;
    }

    eEngineClient=FsmState(pClient);
  }

  switch (eEngineClient) {
  default:
    if (eEngineClient<eEngineClientStateRing) {
DVWRITELN("%ld",dwChunkSizeSource);
      pClient->param.dwChunkSizeSource=dwChunkSizeSource;
DVWRITELN("%ld",pClient->param.dwChunkSizeSource);
      eEngineClient=FsmUp(pClient,eEngineClientStateRing,__FILE__,__LINE__,__func__);

      if (eEngineClient<0) {
        DMESSAGE("creating ring");
        goto exit;
      }
    }

    break;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (RingWrite(pRing,buf,dwChunkSizeSource)<0) {
    DMESSAGE("writing to ring");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  uFramesPlaybackBegin=pClient->param.uFramesPlaybackBegin;
  uFramesWritten=pRing->frames._uWritten;
  uMsLatency=YASAPI_FRAMES2MS(uFramesWritten,&pClient->wfx.source)+0.5;

  if (pClient->stats.uNumFramesMax<uFramesWritten) {
    pClient->stats.uNumFramesMax=uFramesWritten;
    GtkClientUpdateWidget(pGtk,
        &pGtk->config.config.monitor.engine.num_frames_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (UINT32_MAX==pPlayer->stats.uMsLatencyMax
      ||pPlayer->stats.uMsLatencyMax<uMsLatency) {
    pPlayer->stats.uMsLatencyMax=uMsLatency;
    GtkClientUpdateWidget(pGtk,
        &pGtk->config.config.monitor.player.ms_latency);
  }

  /////////////////////////////////////////////////////////////////////////////
  if (eEngineClientStateOpen<=eEngineClient&&eEngineServerStateOpen<=FsmState(pClient->pServer)) {
    goto success;
  }

  uFramesPlaybackBegin=pClient->param.uFramesPlaybackBegin;
  uFramesWritten=pRing->frames._uWritten;
  if (uFramesWritten<uFramesPlaybackBegin) {
    ///////////////////////////////////////////////////////////////////////////
    DVWRITELN("*** underflow: %u (%u %u)",uFramesWritten,uFramesPlaybackBegin,
        dwChunkSizeSource);
    //uFramesMax=pRing->lpVtbl->FramesMaximum(pRing);
    uFramesMax=RingFramesMaximum(pRing);
    pClient->monitor.xWritten=(double)uFramesWritten/uFramesMax;
    GtkClientUpdateWidget(pGtk,&pGtk->config.config.monitor.written);
    goto success;
  } 
  else if (eEngineClient<eEngineClientStateOpen) {
    DVWRITELN("*** starting: %u (%u)",uFramesWritten,uFramesPlaybackBegin);
    eEngineClient=FsmUp(pClient,eEngineClientStateOpen,__FILE__,__LINE__,__func__);

    if (eEngineClient<0) {
      DMESSAGE("opening producer");
      goto exit;
    }
  }
  else if (FsmState(pClient->pServer)<eEngineServerStateOpen) {
    if (EngineClientRequestServerState(pClient,eEngineServerStateOpen )<0) {
      DMESSAGE("requesting eEngineServerStateOpen");
      goto exit;
    }
  }
success:
  code=0;
exit:
ebase:
  return code;
}

int EngineClientWrite(EngineClient *pClient, BYTE *buf, DWORD dwChunkSize)
{
  Player *pPlayer=pClient->pPlayer;
  int code;

  PlayerLock(pPlayer ); // {
  code=_EngineClientWrite(pClient,buf,dwChunkSize);
  PlayerUnlock(pPlayer); // }

  return code;
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientIsPlaying(EngineClient *pClient)
{
  int bIsPlaying=0;

  switch (FsmState(pClient->pServer)) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
    break;
  case eEngineServerStateOpen:
    EngineClientRequestServerState(pClient,eEngineServerStateDraining );

    if (pClient->options.common.bGapless)
      break;
    // Intentional fall-through.
  case eEngineServerStateDraining:
  case eEngineServerStateDrainingFinal:
    bIsPlaying=1;
    break;
  default:
    break;
  }

  return bIsPlaying;
}

int EngineClientIsPlaying(EngineClient *pClient)
{
  Player *pPlayer=pClient->pPlayer;
  int bIsPlaying=0;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto eopen;

  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  bIsPlaying=_EngineClientIsPlaying(pClient);
exit:
  PlayerUnlock(pPlayer); // }
eopen:
  return bIsPlaying;
}

///////////////////////////////////////////////////////////////////////////////
int EngineClientFlush(EngineClient *pClient, int ms)
{
  int code=-1;
  Player *pPlayer=pClient->pPlayer;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DMESSAGE("something went wrong!");
    goto eopen;
  }

  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  if (FsmState(pClient->pServer)<eEngineServerStateOpen)
    goto success;

  EngineClientRequestServerState(pClient,eEngineServerStateFlush );
  pClient->param.time.lpVtbl->Set(pClient->pServer,ms);
success:
  code=0;
exit:
  PlayerUnlock(pPlayer); // }
eopen:
  return code;
}

///////////////////////////////////////////////////////////////////////////////
static int _EngineClientPause(EngineClient *pClient, int nPause)
{
  int nLastPause=pClient->param.nLastPause;
  EngineServer *pServer=pClient->pServer;
  EEngineServer eEngineServer=FsmState(pServer);

  if (nLastPause==nPause)
    goto exit;

  if (nPause) {
    switch (eEngineServer) {
    case eEngineServerStateOpen:
      if (pClient->options.common.bDisconnect) {
DWRITELN("3 {");
        EngineClientRequestServerState(pClient,eEngineServerStatePauseDisconnected );
DWRITELN("3 }");
      }
      else {
DWRITELN("4 {");
        EngineClientRequestServerState(pClient,eEngineServerStatePauseConnected );
DWRITELN("4 }");
      }

      pClient->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
  else {
    switch (eEngineServer) {
    case eEngineServerStatePauseDisconnected:
    case eEngineServerStatePauseConnected:
DWRITELN("5 {");
      EngineClientRequestServerState(pClient,eEngineServerStateOpen );
DWRITELN("5 }");
      pClient->param.nLastPause=nPause;
      break;
    default:
      break;
    }
  }
exit:
  return nLastPause;
}

int EngineClientPause(EngineClient *pClient, int nPause)
{
  Player *pPlayer=pClient->pPlayer;
  int nLastPause=0;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  nLastPause=_EngineClientPause(pClient,nPause);
exit:
  PlayerUnlock(pPlayer); // }
ebase:
  return nLastPause;
}

///////////////////////////////////////////////////////////////////////////////
// Return time history corrected.
int EngineClientOutputTime(EngineClient *pClient)
{
  int ms=0;
  Player *pPlayer=pClient->pPlayer;
  EngineServer *pServer;
  EEngineServer eEngineServer;

  if (pClient->options.common.bLatency)
    goto ebase;
  else if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  pServer=pClient->pServer;
  eEngineServer=FsmState(pServer);

  switch (eEngineServer) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
    goto exit;
  default:
    if (eEngineServer<eEngineServerStateOpen)
      goto exit;

    break;
  }

  ms=pClient->param.time.lpVtbl->Get(pServer);
exit:
  PlayerUnlock(pPlayer); // }
ebase:
  return ms;
}

int EngineClientWrittenTime(EngineClient *pClient)
{
  WAVEFORMATEXTENSIBLE *pTarget=&pClient->wfx.target;
  double msWrittenTime=0.0;
  int nLatency=0;
  Player *pPlayer=pClient->pPlayer;
  Ring *pRing=&pClient->ring;
  EngineServer *pServer;
  EEngineServer eEngineServer;
  WORD uFramesWritten;
  UINT32 uNumFramesPadding;

  if (FsmState(pClient)<eEngineClientStateBase)
    goto ebase;

  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    goto exit;

  pServer=pClient->pServer;
  eEngineServer=FsmState(pServer);

  switch (eEngineServer) {
  case eEngineServerStatePauseDisconnected:
  case eEngineServerStatePauseConnected:
    goto exit;
  default:
    if (eEngineServer<eEngineServerStateOpen)
      goto exit;

    break;
  }

  uFramesWritten=pRing->frames._uWritten;
  nLatency=YASAPI_FRAMES2MS(uFramesWritten,pTarget);

  if (!pClient->param.bDoubleBuffered) {
    if (EngineServerNumFramesPadding(pServer,&uNumFramesPadding)<0)
      goto exit;

    nLatency+=YASAPI_FRAMES2MS(uNumFramesPadding,&pClient->wfx.target);
  }

  if (pClient->param.nMaxLatency<nLatency)
    nLatency=pClient->param.nMaxLatency;

  if (!pClient->options.common.bLatency)
    msWrittenTime+=pClient->param.time.lpVtbl->Get(pServer);
exit:
  PlayerUnlock(pPlayer); // }
ebase:
  return msWrittenTime+nLatency+0.5;
}

void EngineClientVolume(EngineClient *pClient, Player *pPlayer)
{
  PlayerLock(pPlayer ); // {

  if (FsmState(pClient)<eEngineClientStateOpen)
    gcIEngineServerVolumeNullVtbl.Set(pClient->pServer);
  else
    pClient->param.volume.lpVtbl->Set(pClient->pServer);

  PlayerUnlock(pPlayer); // }

  return;
}

void EngineClientSetActive(EngineClient *pClient, int bActive)
{
  Player *pPlayer=pClient->pPlayer;

  if (eEngineClientStateBase<FsmState(pClient)) {
    PlayerLock(pPlayer ); // {
    pClient->param.bActive=bActive;
    DVWRITELN("==> bActive: %d <==================================",bActive);
    PlayerUnlock(pPlayer); // }
  }
}


static int EngineClientPreOpen(EngineClient *pClient, int bIncBroken,
    int bIncTracks)
{
  Player *pPlayer=pClient->pPlayer;
  EngineParam *pParam=&pClient->param;
  EngineOptions *pOptions=&pClient->options;

  if (FsmState(pClient)<eEngineClientStateBase) {
    DVMESSAGE("unexpected state %s",FsmStateId(pClient));
    goto exit;
  }

  FsmDown(pClient,eEngineClientStateBase );

  if (FsmUp(pClient,eEngineClientStateClose,__FILE__,__LINE__,__func__)<0) {
    DMESSAGE("entering eEngineClientStateClose failed");
    goto exit;
  }

  if (EngineParamCreate(pParam,pOptions,pClient)<0) {
    DMESSAGE("creating parameter");
    goto exit;
  }

  // During "open" we thought this is "gapless". Meanwhile the sesssion was
  // terminated.
  if (FsmUp(pClient,eEngineClientStatePreOpen,__FILE__,__LINE__,__func__)<0) {
    DMESSAGE("creating gapless sequence");
    goto exit;
  }

  if (bIncBroken)
    YASAPI_INC_UINT32(pClient->pPlayer->stats.uNumGaplessBroken);

  if (bIncTracks) {
    YASAPI_INC_UINT32(pClient->pPlayer->stats.uNumTracks);
    YASAPI_INC_UINT32(pClient->stats.uNumTracks);
  }

  if (UINT32_MAX==pClient->pPlayer->stats.uNumGaplessBroken)
    pClient->pPlayer->stats.uNumGaplessBroken=0lu;

  pClient->stats.uNumFramesDropout=UINT32_MAX;
  YASAPI_INC_UINT32(pClient->pPlayer->stats.uNumGapless);
  GtkClientUpdateDialog(&pPlayer->gtk,&pPlayer->gtk.config);

  return 0;
exit:
  return -1;
}

static EEngineServer EngineClientServerState(EngineClient *pClient)
{
  if (FsmState(pClient)<eEngineClientStateBase)
    return eEngineServerStateStart;
  else
    return FsmState(pClient->pServer);
}
